use prost::Message;
use std::sync::Mutex;
use std::time::Duration;
use std::{
    io::{Cursor, Write},
    net::{TcpListener, TcpStream},
    sync::{
        atomic::{AtomicBool, Ordering},
        Arc,
    },
    thread,
};

use commander_kernel::router::Action;
use commander_kernel::tcp::ConnectOptions;
use commander_kernel::{
    router::Router,
    stream::{StreamReader, StreamReaderOptions},
    Payload,
};
use errors::HostError;

pub struct Host {
    connect_options: ConnectOptions,
    should_stop: Arc<AtomicBool>,
    handle: Option<thread::JoinHandle<Result<(), HostError>>>,
    router: Arc<Mutex<Router>>,
}

impl Host {
    pub fn new(options: ConnectOptions) -> Host {
        Host {
            connect_options: options,
            should_stop: Arc::new(AtomicBool::new(false)),
            handle: None,
            router: Arc::new(Mutex::new(Router::new())),
        }
    }

    pub fn add_action(&mut self, route: &str, action: Box<dyn Action>) {
        let router = self.router.clone();
        let mut router = router.lock().unwrap();
        router.add_action(route, action);
    }

    pub fn start(&mut self) -> Result<(), HostError> {
        let should_stop = self.should_stop.clone();
        let options = self.connect_options.clone();

        if should_stop.load(Ordering::Relaxed) {
            return Ok(());
        }

        let router = self.router.clone();
        self.handle = Some(thread::spawn(move || -> Result<(), HostError> {
            let listener = match TcpListener::bind(format!("{}:{}", options.ip, options.port)) {
                Ok(it) => it,
                Err(_err) => return Err(HostError::BindError),
            };

            for mut stream in listener.incoming() {
                println!("received data");
                if should_stop.load(Ordering::Relaxed) {
                    break;
                }

                let mut reader = StreamReader::new(StreamReaderOptions {
                    limit: 1024,
                    read_size: 128,
                });

                let mut stream = match stream {
                    Ok(s) => s,
                    Err(e) => return Err(HostError::ReadError),
                };

                match stream.set_read_timeout(Some(Duration::from_millis(10))) {
                    Ok(_) => {}
                    Err(_) => return Err(HostError::ReadError),
                };

                let bytes = reader.read(&mut stream);

                let bytes = match bytes {
                    Ok(b) => b,
                    Err(e) => return Err(HostError::ReadError),
                };

                let payload = Payload::decode(Cursor::new(&bytes));
                let payload = match payload {
                    Ok(p) => p,
                    Err(e) => return Err(HostError::DecodeError),
                };

                let router = router.lock().unwrap(); // TODO: do we want to handle a potential deadlock? Is this even the correct rust pattern for locking?
                let return_data = router.call(payload.route, &payload.body.to_vec());

                match stream.write(&return_data) {
                    Ok(_) => {} // do nothing for now
                    Err(_) => return Err(HostError::WriteError),
                }
            }
            Ok(())
        }));

        Ok(())
    }

    pub fn stop(&mut self) -> Result<(), HostError> {
        self.should_stop.store(true, Ordering::Relaxed);

        if self.handle.is_some() {
            match TcpStream::connect(format!(
                "{}:{}",
                self.connect_options.ip, self.connect_options.port
            )) {
                Ok(mut stream) => match stream.write(b"") {
                    Ok(_) => {
                        println!("empty message written to server to signal a potential stop");
                    }
                    Err(_e) => {
                        println!("error writing stop message to server");
                        return Err(HostError::WriteError);
                    }
                },
                Err(_e) => return Err(HostError::ConnectError),
            }

            self.handle
                .take()
                .expect("no handle")
                .join()
                .expect("couldn't join");
        }

        Ok(())
    }
}

mod errors {
    pub enum HostError {
        BindError,
        ConnectError,
        WriteError,
        ReadError,
        DecodeError,
    }
}
