use std::collections::LinkedList;

/// a dynamic buffer of `u8`, internally the `Vec<u8>` is added to a `LinkedList<Vec<u8>>`
pub struct DynamicBuffer {
    list: LinkedList<Vec<u8>>,
}

impl DynamicBuffer {
    /// Returns a new buffer
    pub fn new() -> DynamicBuffer {
        DynamicBuffer {
            list: LinkedList::new(),
        }
    }

    /// Adds a given byte buffer to the dynamic buffer.
    ///
    /// If you have a buffer of length 256, but it is only filled with 100, then this will slice it to `len`, so you only end up with a buffer that is filled with exactly the content, and does not fill up memory with wasted space.
    ///
    /// # Arguments
    /// * `buffer` - a given bytebuffer to be added to the dynamic buffer
    /// * `len` - the length of the content that is actually filled with contents
    ///
    /// # Examples
    ///
    /// ```
    /// use commander_kernel::buffer::DynamicBuffer;
    ///
    /// let mut buffer = DynamicBuffer::new();
    /// buffer.add(vec![0, 0, 0, 0], 4);
    /// ```
    pub fn add(&mut self, buffer: Vec<u8>, len: usize) {
        self.list.push_back(buffer[..len].to_vec());
    }

    /// Returns the length of the buffer. Here the length is how many `Vec<u8>` that has been added to the buffer, and not the length of all the added `Vec<u8>`.
    pub fn len(&self) -> usize {
        self.list.len()
    }

    /// merges all the added `Vec<u8>` to a single `Vec<u8>`
    ///
    /// # Examples
    ///
    /// ```
    /// use kernel::buffer::DynamicBuffer;
    ///
    /// let mut buffer = DynamicBuffer::new();
    ///
    /// buffer.add(vec![0, 0, 0, 0], 4);
    /// buffer.add(vec![1, 1, 1], 3);
    ///
    /// let merged = buffer.merge();
    ///
    /// assert_eq!(7, merged.len());
    /// ```
    pub fn merge(&self) -> Vec<u8> {
        let mut merged = vec![];

        for byte_buffer in &self.list {
            merged.extend(byte_buffer.iter());
        }

        merged
    }

    #[must_use]
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }
}

mod tests {
    use super::DynamicBuffer;

    #[test]
    pub fn add_vec_to_buffer() {
        let mut buffer = DynamicBuffer::new();
        let vector = vec![0, 0, 0, 0];
        buffer.add(vector, 4);

        assert_eq!(1, buffer.len());
    }

    #[test]
    pub fn odd_buffer_lengths() {
        let test = vec![1, 2, 3, 4, 5, 0, 0, 0];
        let mut buffer = DynamicBuffer::new();

        buffer.add(test, 2);

        let merged = buffer.merge();

        assert_eq!([1u8, 2u8], merged.as_slice());
        assert_eq!(2, merged.len());
    }

    #[test]
    pub fn add_many_to_buffer() {
        let mut buffer = DynamicBuffer::new();
        let test = vec![0, 1, 2, 3, 4, 5, 6, 7];
        buffer.add(test[0..3].to_vec(), 3);
        buffer.add(test[3..].to_vec(), 5);

        let merged = buffer.merge();

        assert_eq!(2, buffer.len());
        assert_eq!(8, merged.len());

        let mut counter = 0;
        loop {
            assert_eq!(test[counter], merged[counter]);

            if counter == test.len() - 1 {
                break;
            }

            counter += 1;
        }
    }
}
