use crate::crypto::errors::CryptoError;
use rsa::signature::SignatureEncoding;
use rsa::{
    pkcs1::DecodeRsaPrivateKey,
    pkcs1::DecodeRsaPublicKey,
    pkcs1v15::{SigningKey, VerifyingKey},
    sha2::Sha256,
    signature::RandomizedSigner,
    signature::Verifier,
    Pkcs1v15Encrypt, RsaPrivateKey, RsaPublicKey,
};

pub trait Encrypt {
    fn encrypt(&self, data: &[u8]) -> Result<Vec<u8>, CryptoError>;
}

pub trait Decrypt {
    fn decrypt(&self, data: &[u8]) -> Result<Vec<u8>, CryptoError>;
}

pub trait Signature {
    fn sign(&self, data: &[u8]) -> Vec<u8>;
    fn verify(&self, data: &[u8], signature: &[u8]) -> Result<bool, CryptoError>;
}

// TODO: This is an UGLY name. Rename it something other than Handler. Almost everything is a handler
struct RSACryptoHandler {
    private_key: RsaPrivateKey,
    public_key: RsaPublicKey,
}

impl RSACryptoHandler {
    fn new_from_str(private_key: &str, public_key: &str) -> Result<RSACryptoHandler, CryptoError> {
        let private_key = match RsaPrivateKey::from_pkcs1_pem(private_key) {
            Ok(k) => k,
            Err(_) => return Err(CryptoError::PrivateKeyInvalid),
        };

        let public_key = match RsaPublicKey::from_pkcs1_pem(public_key) {
            Ok(k) => k,
            Err(_) => return Err(CryptoError::PublicKeyInvalid),
        };

        Ok(RSACryptoHandler {
            private_key,
            public_key,
        })
    }

    fn new_from_file(
        private_pem_key_path: &str,
        public_pem_key_path: &str,
    ) -> Result<RSACryptoHandler, CryptoError> {
        let private_key = match RsaPrivateKey::read_pkcs1_pem_file(private_pem_key_path) {
            Ok(k) => k,
            Err(_) => return Err(CryptoError::PrivateKeyNotFound),
        };

        let public_key = match RsaPublicKey::read_pkcs1_pem_file(public_pem_key_path) {
            Ok(k) => k,
            Err(_) => return Err(CryptoError::PublicKeyNotFound),
        };

        Ok(RSACryptoHandler {
            private_key,
            public_key,
        })
    }
}

impl Encrypt for RSACryptoHandler {
    fn encrypt(&self, data: &[u8]) -> Result<Vec<u8>, CryptoError> {
        // TODO: should this be called every time?
        let mut rng = rand::thread_rng();
        let encrypted_data = self.public_key.encrypt(&mut rng, Pkcs1v15Encrypt, data);

        match encrypted_data {
            Ok(d) => Ok(d),
            Err(_) => Err(CryptoError::Encrypt),
        }
    }
}

impl Decrypt for RSACryptoHandler {
    fn decrypt(&self, data: &[u8]) -> Result<Vec<u8>, CryptoError> {
        let decrypted_data = self.private_key.decrypt(Pkcs1v15Encrypt, data);

        match decrypted_data {
            Ok(d) => Ok(d),
            Err(_) => Err(CryptoError::Decrypt),
        }
    }
}

impl Signature for RSACryptoHandler {
    fn sign(&self, data: &[u8]) -> Vec<u8> {
        // TODO: should this be called every time?
        let mut rng = rand::thread_rng();

        // TODO: is it stupid to use clone here?
        let signing_key = SigningKey::<Sha256>::new(self.private_key.clone());
        let signature = signing_key.sign_with_rng(&mut rng, data);

        signature.to_vec()
    }

    fn verify(&self, data: &[u8], signature: &[u8]) -> Result<bool, CryptoError> {
        let signature = match rsa::pkcs1v15::Signature::try_from(signature) {
            Ok(s) => s,
            Err(_) => return Err(CryptoError::InvalidSignature),
        };

        // TODO: is it stupid to use clone here?
        let verifying_key = VerifyingKey::<Sha256>::new(self.public_key.clone());

        match verifying_key.verify(data, &signature) {
            Ok(_) => Ok(true),
            Err(_) => Err(CryptoError::SignatureVerification),
        }
    }
}

mod errors {
    #[derive(Debug)]
    pub enum CryptoError {
        Verify,
        Sign,
        Encrypt,
        Decrypt,
        PrivateKeyNotFound,
        PublicKeyNotFound,
        PrivateKeyInvalid,
        PublicKeyInvalid,
        InvalidSignature,
        SignatureVerification,
    }
}

mod tests {
    use crate::crypto::{Decrypt, Encrypt, RSACryptoHandler, Signature};

    fn get_public_key() -> String {
        String::from(
            "-----BEGIN RSA PUBLIC KEY-----
MIIBCgKCAQEAnFkuwQIuo0c8Ab2T0iiJ6X17epAfsiL7TryiAdwh23eIOezQgegl
Sd6rl21GIMqb1ECKxtjGmzVv7jrh4tN/yoRky944vWFFeg85PJ7nRwiYLbV4mmgk
QDkTjbzeOLhGyDISUQgqPxXPPZbupr1/QWasiSCpA7D1G+KLS7oJq/7RKpK9AaOB
XxjQPBjARTouO/I57pDICrcnxZ6tpba1lvERaabsO0ZTVBm2PDefyIYQ4mpyWOJX
j4ES8fIfnQPfejxolkROkcMcuWOyOo7oh+pcV1q8yDOiGMo6eGHvbJWEKnfGwNn1
iHGZiJGdDf3AwIdN4KRBprS45bCoGS1+3QIDAQAB
-----END RSA PUBLIC KEY-----
",
        )
    }
    fn get_private_key() -> String {
        String::from(
            "-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAnFkuwQIuo0c8Ab2T0iiJ6X17epAfsiL7TryiAdwh23eIOezQ
geglSd6rl21GIMqb1ECKxtjGmzVv7jrh4tN/yoRky944vWFFeg85PJ7nRwiYLbV4
mmgkQDkTjbzeOLhGyDISUQgqPxXPPZbupr1/QWasiSCpA7D1G+KLS7oJq/7RKpK9
AaOBXxjQPBjARTouO/I57pDICrcnxZ6tpba1lvERaabsO0ZTVBm2PDefyIYQ4mpy
WOJXj4ES8fIfnQPfejxolkROkcMcuWOyOo7oh+pcV1q8yDOiGMo6eGHvbJWEKnfG
wNn1iHGZiJGdDf3AwIdN4KRBprS45bCoGS1+3QIDAQABAoIBAALYcawkFsa0d3Bd
q6jNwZLNDJgCsvz6O4Inlzg68ifNIV7Sg0n1+GXDhnJzU+0pWtQuV4g+bkDpZE/c
5PsMl90O80mtkfFKaaoKyGMkzjdB/gkI7ji7L/u2Pjecahv0qJ8goCo24JP02Oh5
TH8e4EfHr1iwjjrpt9BJ23ilmmL1sT3rqhlmzPHhAMzYrce20Gj/wBWeZ4xUmTYf
tw0C4eQkJOI7usFSz+gHexEkD5xhYDQPHIaySp7vzQhwVKO0Nqq47oIQ73ju0Va7
GjkIA9QOvekfLiXKS4rbiKwTfZUNPXA2aT/oA4+4qj24/I5FdvZxU2TKl2JWCmJk
WmqBhMkCgYEAz0ixkJz+wSHqWW2MjcssZ8qXj53hQ/9ucKnlFl37VlvAaOcH/vwN
zAmjhuhtrMXvZodNHnqUVv/9/NW16/IuiNPIbR0VyqGswQPzhT96We1tIZUFguky
NhGC/MRPrQEFikqHiz5znkCWc/qVi70MhEgjyKBQhAnCUNYKPr+nWNkCgYEAwRfu
zFT/QB63F2HlS47MjftF74H/ziX5WMjJRmxk81/iXdhlFh9jMMZ9CErM9CZGDcwa
dymgFXAjotstiVnXkO+e6Zsa+cxKBpj1atG2hooODtfGwt12N3cHz7xMtk+y2wgp
Ap4xAyYcS4O9qGn++dj2yrZuNbbTc9OcJoXOM6UCgYEAiJ3GuiNfHHd393zmWEYC
fF3gSn2qDKD7bzHqlTVYA/Dw/bLXaE1D4qivqKjkMC08vfyK7X2o6zZJdix8fIff
97ptrMywoh6wt34dMENIJqD/flSzjzaXK1+bT3ErekDkH+NDNpjwpuq2vrX3fmQZ
Lc5GF2Lh7kRoV9LaacTzTPECgYBc5dDDiZkMFjrsCvMv+NG89UjeQ+L4IRjneCIt
uXSQCm9zyupHTG6CRtxcovAb/lXxyWJqNiKIbd0XOu+bhxfI6R/BMBg4Weq7Zdec
vqxb6gt9ZJ6Bnd6ELMYbUXJuu8VGz5Q7aDQYWbt/MTCL2AMN3S5OtJotSXpN3V7g
wbtE6QKBgQCGF2diTGgHg9H7KUSmBWDVQVM1RhMg3GfTdEITGBPcMRdrx1y/NHYE
99xaS3XZgWqz05Xn+Ku5Dntiu9QwDffA02ZhwGcgn7ZLaWCU3sCHuK4mTNmK2QpG
m/A1vde9ULfntRXDT+44rL5BfsLB9oNfknDtLSk/TsE9zQ6f/zghzw==
-----END RSA PRIVATE KEY-----
",
        )
    }
    #[test]
    fn test_rsa_encrypt() {
        let data = b"hello from test";
        let private_key = get_private_key();
        let public_key = get_public_key();
        let crypto_handler =
            RSACryptoHandler::new_from_str(private_key.as_str(), public_key.as_str()).unwrap();

        let encrypted_data = crypto_handler.encrypt(data).unwrap();
        let encrypted_data = &encrypted_data[..];

        assert!(data.iter().ne(encrypted_data.iter()))
    }

    #[test]
    fn test_rsa_decrypt() {
        let data = b"hello from test";
        let private_key = get_private_key();
        let public_key = get_public_key();
        let crypto_handler =
            RSACryptoHandler::new_from_str(private_key.as_str(), public_key.as_str()).unwrap();

        let encrypted_data = crypto_handler.encrypt(data).unwrap();
        let encrypted_data = &encrypted_data[..];
        let decrypted_data = crypto_handler.decrypt(encrypted_data).unwrap();
        let decrypted_data = &decrypted_data[..];

        assert!(data.iter().ne(encrypted_data.iter()));
        assert!(data.iter().eq(decrypted_data.iter()));
    }

    #[test]
    fn test_sign_and_verify() {
        let data = b"Hello from test";
        let private_key = get_private_key();
        let public_key = get_public_key();
        let crypto_handler =
            RSACryptoHandler::new_from_str(private_key.as_str(), public_key.as_str()).unwrap();

        let signature = crypto_handler.sign(data);
        let verified = crypto_handler.verify(data, &signature[..]).unwrap();

        assert!(data.iter().ne(signature.iter()));
        assert!(verified);
    }
}
