use std::collections::HashMap;

pub struct Router {
    router_map: HashMap<String, Box<dyn Action>>,
}

impl Router {
    pub fn new() -> Self {
        Self {
            router_map: HashMap::new(),
        }
    }

    pub fn add_action(&mut self, route: &str, action: Box<dyn Action>) {
        self.router_map.insert(route.to_string(), action);
    }

    pub fn call(&self, route: String, data: &Vec<u8>) -> Vec<u8> {
        let action = self.router_map.get(&route);

        let action = match action {
            None => return [0u8; 0].to_vec(), // TODO: we need to make the return of call an Option, with either something or nothing, so the caller can decide what to do
            Some(a) => a,
        };

        action.work(&data)
    }
}

pub trait Action: Send + Sync {
    // TODO: this always needs to return, which isn't always the case: something could go wrong. Make it more "errorable"
    // TODO: the return type should maybe just be an array?
    fn work(&self, data: &Vec<u8>) -> Vec<u8>;
}
mod tests {
    use super::{Action, Router};

    struct TestAction;

    impl Action for TestAction {
        fn work(&self, data: &Vec<u8>) -> Vec<u8> {
            [0u8; 1].to_vec()
        }
    }
    #[test]
    fn create_router_with_one_action() {
        let mut router = Router::new();
        let action = TestAction;
        router.add_action("my/action", Box::new(action));
    }

    #[test]
    fn call_action_in_router_once() {
        let mut router = Router::new();
        let action = TestAction;
        router.add_action("my/action", Box::new(action));

        router.call("my/action".to_string(), &vec![0u8, 1u8, 2u8]);
    }

    #[test]
    fn call_non_exsistant_action() {
        let mut router = Router::new();
        let action = TestAction;
        router.add_action("my/action", Box::new(action));

        router.call("my/action/none".to_string(), &vec![0u8, 1u8, 2u8]);
    }

    #[test]
    fn call_action_in_router_twice() {
        let mut router = Router::new();
        let action = TestAction;
        router.add_action("my/action", Box::new(action));

        router.call("my/action".to_string(), &vec![0u8, 1u8, 2u8]);
        router.call("my/action".to_string(), &vec![0u8, 1u8, 2u8]);
    }
}
