use std::io::Read;

use crate::buffer::DynamicBuffer;

use self::errors::StreamReaderError;

pub struct StreamReaderOptions {
    pub limit: usize,
    pub read_size: usize,
}

pub struct StreamReader {
    limit: usize,
    read_size: usize,
    dynamic_buffer: DynamicBuffer,
}

impl StreamReader {
    pub fn new(options: StreamReaderOptions) -> StreamReader {
        StreamReader {
            read_size: options.read_size,
            limit: options.limit,
            dynamic_buffer: DynamicBuffer::new(),
        }
    }

    pub fn read<T: Read>(&mut self, stream: &mut T) -> Result<Vec<u8>, StreamReaderError> {
        let mut total_size = 0;

        loop {
            let mut buffer = vec![0u8; self.read_size];
            let read_bytes_len = match stream.read(&mut buffer) {
                Ok(len) => len,
                Err(_) => break, // we return the buffer read - the read_timeout is set to 1 sec
            };

            total_size += read_bytes_len;

            if total_size > self.limit {
                return Err(StreamReaderError::LimitReached);
            }

            // we are done reading, because client closed connection (on Linux),
            // according to https://doc.rust-lang.org/stable/std/io/trait.Read.html#tymethod.read
            if read_bytes_len == 0 {
                break;
            }

            self.dynamic_buffer.add(buffer, read_bytes_len);

            // no more reading to be done
            // TODO: this doesn't handle the edge case where read_size == read_bytes_len: stream.read(...) will still block :(
            if read_bytes_len < self.read_size {
                break;
            }
        }

        Ok(self.dynamic_buffer.merge())
    }
}

mod errors {
    pub enum StreamReaderError {
        ReadError,
        LimitReached,
    }
}
